#!/usr/bin/python

# v 0.1.10


# OPTIONS:

# HOST is the address of the server. It can be the same as HOST_URI.
HOST="192.168.0.2"
PORT=80

# HOST_URI is passed in the header "Host:" during the HTTP requests.
HOST_URI="httpvpn.example.org"

# URL where the mirror script resides. Remember the / at the beginning 
URL="/mirror.php"

# MODE indicates which type of script is in use: php or cgi.
MODE="php"

# ME is the localhost "name". This name is used to created the pipe on the server and 
# as an address to this host, so it must be unique on this server.
# It can be any combination of letters that can be used for an URL.
ME="B"

# PEER is the name of the client we want to connect to. See ME for details.
PEER="A"

# L_IP and L_NETMASK are the local IP and netmask of the tunnel interface.
L_IP="10.0.0.1"
L_NETMASK="255.255.255.0"
# MTU is the MTU to be used in the tunnel interface. 
MTU=2000

# KEY is the Shared key you are using in the Encrypt/Decrypt.
# When using Blowfish encryption, the key size must be a multiple of 8
DISABLE_ENCRYPTION=0
KEY="HTTPVPN12345!!!!"

# PROXY specifies the IP address of the proxy. In order for this option to 
# have effect, PROXY_ENABLE must be set to 1

PROXY_ENABLE=0
PROXY="194.176.176.82"
PROXY_PORT=8080

# TIMEOUT is the timespan (in seconds) before sending a keep-alive to the http server.
TIMEOUT=9

# DEBUG specifies the output level of debugging information (0=none,1=normal,2,3..)
DEBUG=0

# Trace enables live debugging
TRACE=0

### CRYPTOGRAPHIC CODE ###
# Note: This code was intentionally separated from the VPN itself to give the ability to
#       anyone to change and experiment with different encryption functions.


from Crypto.Cipher import Blowfish
from Crypto.Hash import SHA

def EncryptData(s,key):

  # Save data length
  data_len = hex(len(s))[2:].zfill(4)

  s = "%s%s" % (data_len , s)

  # Padding data to be multiple of 8
  l=len(s)+40
  pad = (8 - (l%8) ) + l
  s = s.ljust(8-len(s)%8+len(s),"X")


  # Calculate the digest
  digest = SHA.new(s).hexdigest()
 
  # put all together
  s = "%s%s" % (digest, s)

  blowfish = Blowfish.new(key)
  
  data=blowfish.encrypt(s)
  return (data,0)  

def DecryptData(s,key):
  blowfish= Blowfish.new(key)

  try:
    data = blowfish.decrypt(s)
  except:
    print "Error: DecryptData "
    return("",1)
   
  digest = data[0:40]
  data=data[40:]
  new_digest = SHA.new(data).hexdigest()

  if( new_digest == digest):
     l = int(data[0:4],16)
     data = data[4:l+4]

     return (data,0)
  else:
    print "Digest don't match: " + new_digest + " vs " + digest
    return ("",1)


### CODE ###
import os, sys
from socket import *
from fcntl import ioctl
from select import select
import getopt, struct
from base64 import b64encode,b64decode
from time import time
import httplib

from md5 import md5

if(TRACE):
  import pdb

# This are just constant needed by the TUN/TAP driver
TUNSETIFF = 0x400454ca
IFF_TUN   = 0x0001
IFF_TAP   = 0x0002

TUNMODE = IFF_TUN
#TUNMODE = IFF_TAP

"""
   Connect the Receiver to the webserver
"""
def Recv_Connect(url):

  recv_s = socket(AF_INET, SOCK_STREAM)
  if (PROXY_ENABLE):
    recv_s.connect((PROXY,PROXY_PORT))
  else:
    recv_s.connect((HOST,PORT))

  recv_s.send("GET " + url + " HTTP/1.1\r\n")

  if(PROXY_ENABLE==0):
    recv_s.send("Host: "+HOST_URI +"\r\n")

  recv_s.send("\r\n")
  # Convert the socket to a file to make it easier to use, also note that
  # the buffer's lenght equals the MTU + headers + base64 encoding ratio 
  recv_f=recv_s.makefile("r+",((MTU+20+40+8)*8/6)+1 )

  return recv_f


"""
 Main
"""

# Create the tun device
f = os.open("/dev/net/tun", os.O_RDWR)
ifs = ioctl(f, TUNSETIFF, struct.pack("16sH", "tun%d", TUNMODE))

ifname = ifs[:16].strip("\x00")
print "Allocated interface %s." % ifname

# Configure the interface
if (L_IP):
  os.system("ifconfig " + ifname + " " + L_IP + " netmask " + L_NETMASK + " mtu " + str(MTU) )
  print L_IP + "/" + L_NETMASK + "(mtu " + str(MTU) + ") assigned."


"""
 Preparing URLS
"""
if (MODE=="php"):
  recv_url=URL+"?s="+ PEER+ "&d=" + ME +"&m=r"
  send_url=URL+"?s="+ ME + "&d=" + PEER +"&m=w"
else:  
  recv_url=URL+"?"+ PEER+ ":" + ME +":r"
  send_url=URL+"?"+ ME + ":" + PEER +":w"

if (PROXY_ENABLE):
  recv_url="http://"+HOST_URI+recv_url
  send_url="http://"+HOST_URI+send_url 

recv_f=Recv_Connect(recv_url)

if(DEBUG):
  print "Recv_url:"+recv_url
  print "Send_url:"+send_url

"""
 Connect to send data
"""
if(PROXY_ENABLE):
  print "Using Proxy " + PROXY + ":" + str(PROXY_PORT)
  send_h = httplib.HTTPConnection(PROXY,PROXY_PORT)
else:
  send_h = httplib.HTTPConnection(HOST_URI)

if(DEBUG >=3 ):
  send_h.set_debuglevel(2)

# recv_status=0 means that we are waiting for the webserver acknowledge
recv_status=0

try: 
  pid = os.fork() 
  if pid > 0:
  # Parent process: Receiver
    while 1:
      r = select([recv_f],[],[],TIMEOUT)[0]
      if ( not r ):
      # Timeout: We didn't received any packet in TIMEOUT seconds. 
      #          Normally the keep_alive on mirror.php is shorter than this timeout. 
      #          This probably means we lost the connection with the webserver
        print "recv: Timeout. Connecting again to " + recv_url
        recv_f=Recv_Connect(recv_url)
        recv_status=0
     
      elif (r[0]==recv_f):
        try:
        # Data from HTTP
          data=recv_f.readline()
          if ( recv_status ):
            # We need to know when we received the last packet to know if 
            # we lost the connection with the webserver
            last_pkt_timestamp=time()

            if(int(data,16)==1):
              # Skip keep-alive packets
              if (DEBUG):
                print "skip keep-alive:" + data
              recv_f.read(3)	# \n\r\n
            else:
              # Get chunk size
              # Most server limit chunk size to 0x2000 (8192 byte)
       
              c_len = int(data,16)
              data=recv_f.read(c_len)
              if(c_len==8192):
                while(data[-1]!='\n'):
                  if(DEBUG):
                    print "Fetching more chunks"
                  recv_f.read(2)
                  c_len=int(recv_f.readline()[:-1],16)
                  data="%s%s"%(data,recv_f.read(c_len))

              # Get rid of CRLF
              recv_f.read(2)
         
              if(DEBUG):
                print "Data from HTTP length=%s:%s"%(str(len(data)),md5(data).hexdigest())

              clear_data=""
              try:
                clear_data=b64decode(data[:-1])    
              except:
                if(TRACE):
                  pdb.set_trace()

              err=0
              if(not DISABLE_ENCRYPTION):
                (clear_data,err)=DecryptData(clear_data,KEY)
              if(not err):
                if(DEBUG>=2):
                  print "Decoded %s:%s" % (str(len(clear_data)),md5(clear_data).hexdigest())
                # Writes data to tun/tap
                os.write(f,clear_data)
              else:
                print "An error has occured while Decrypting Data: "+ str(err)   

          else:
            # We are waiting for the server ack
            if( data[:15] =="HTTP/1.0 200 OK"):
              print "WARNING: HTTP/1.0 The system MAY not work"

            # Check if answer is HTTP/1.X 200 OK
            if( data[:7] == "HTTP/1." and data[9:15] =="200 OK" ):
              if(DEBUG):
                print "Waiting for recv_ack"
                print data

              d=recv_f.readline()
              # Discard Headers
              while (d!="\r\n"):
                if(DEBUG >=3):
                  print len(d)
                  print d
                d=recv_f.readline()
         
              # Read chunk size and convert to decimal
              c_len = int(recv_f.readline(), 16)
              d=recv_f.read(c_len)
          
              # Get rid of ending CRLF
              recv_f.read(2)
              recv_status=1
              print "recv: Connected"
        except:
          # Reset the connection
          print "Error: HTTP data"
          recv_f=Recv_Connect(recv_url)
          recv_status=0


  else:
  # Child process: Sender
    while 1:
      r = select([f],[],[],TIMEOUT)[0]
    
      # Data from TUN/TAP
      data = os.read(f,MTU + 20)  
      if (DEBUG):
        print "Data From TUN length=%s:%s" % (str(len(data)),md5(data).hexdigest())

      err=0

      if(not DISABLE_ENCRYPTION):
        (data,err) = EncryptData(data,KEY)

      if (not err):
        encoded_data = b64encode(data)+"\n"
        if(DEBUG >= 2):
          print "Encoded %s:%s" % (str(len(encoded_data)),md5(encoded_data).hexdigest())

        send_h.request("POST",send_url,encoded_data)
        try:
           res = send_h.getresponse()  
        except httplib.BadStatusLine:
           if(DEBUG):
             print "BadStatusLine reconnecting..."
           if(PROXY_ENABLE):
             send_h = httplib.HTTPConnection(PROXY,PROXY_PORT)
           else:
             send_h = httplib.HTTPConnection(HOST_URI)  
           send_h.request("POST",send_url,encoded_data)
           res = send_h.getresponse()
               
        if (res.status == 200):
          res.read()
        else:
          print "An error has occured while sending data: " + str(res.status)
          print res.read()
      else:
        print "An error has occured while Encrypting the data: "+ str(err)
   
except OSError, e: 
  print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
  sys.exit(1)



